<!---
Memory-sensitive cache implementation for CF.

For those of you haven't read my blog entry, this is a memory-sensitive
cache implementation for CF that makes liberal use of a couple of Java
classes, SoftReference and ReferenceQueue. Details aside, this cache
will keep storing data as long as there's enough memory available. When
memory becomes constrained, either because the cache is too large, or
because some other objects in the JVM need memory, the cache will automatically
shrink.

So, don't assume that just because you put something in here, it'll stay in!
It's entirely possible that an entry that you put into the cache will be gone
when you look in on it the next time, so build logic around this so that if an
entry is not available, it's retrieved From underlying storage, or wherever,
and stuck in the cache again.

In theory the cache will shrink only when the garbage collector decides that
the JVM needs memory; however, I have seen cases Where items are cleaned out
of the cache a few minutes after they're put in, so watch out for that. Your
mileage may vary... The cache is instrumented to tell you its hit ratio, so
watch that, and if it drops too low, you might want to think about adopting
alternate, or supplementary, caching strategies.

I have not tested this extensively; which is not to say that I haven't tested it!
If you do find any bugs, leave a comment on the blog entry below, and I'll
see what I can do about it.

Usage: 

Set the cache into a shared scope, probably server or application:
<cfset server.mycache = createObject("component", "softcache")>
Then get/set/delete entries as you will. When you get an entry, be sure to make the
comparison against server.mycache.notfound, which could vary based on what you choose 
to stick in there, simple type, CFC, whatever. 

An optional add-on is to specify the onRemoveHandler for this object, which if available 
will be invoked when a key is  being removed From the cache due to memory constraints; note 
that the value may already have been cleared by the garbage collector. The onRemoveHandler 
must have a public function named onRemove which takes the key string as an argument. Here's 
how to add it on:
<cfset server.cache.onRemoveHandler = createObject("component", "myOnRemoveHandler")>

The reap() function clears out keys for garbage collected entries. Under normal usage, you'll
never need to call this function, since it is invoked when the get() function is called and a 
cache miss occurs. However, this means that callers pay the cost of reaping the cache; if you
want to (typically you won't need to), you can call this function manually, perhaps as a scheduled
task.

Ashwin Mathew - for more see http://blogs.sanmathi.org/ashwin/
--->
<cfcomponent name="softcache" output="false">

	<!--- 
	Just a constant to indicate a key is not found. 
	If you need to, you can change this to something you can be sure
	will not be a value set by your apps. It could even be a CFC, if
	CFCs are what you're storing in the cache. Your "notfound" CFC
	could implement a function that tells you that it's "not found".
	Not that I've tried it yet - just an idea...
	--->
	<cfset this.notfound = "notfound">

	<!---
	The Struct containing mappings of keys to values wrapped
	in soft references.
	--->
	<cfset this.cache = StructNew()>
	
	<!---
	A Struct mapping soft reference objects to keys. We need
	this to clean out keys From the cache - you'll see why later.
	--->
	<cfset this.refKeyMap = StructNew()>
	
	<!---
	The reference queue that gets garbage collected soft references.
	--->
	<cfset this.refq = createObject("java", "java.lang.ref.ReferenceQueue")>
	
	<!---
	Stats for calculating the hit ratio.
	--->
	<cfset this.hits = 0>
	<cfset this.misses = 0>


	<cffunction name="cacheInit" output="No" returntype="any">
		<cfargument name="scope" default="application">
		<cfset this.scope = scope>
		<cfreturn this>
	</cffunction>

	<cffunction name="cacheGetElement" output="false" description="Returns the value for the specified key, or 
								the string specified by this.notfound if unavailable">
		<cfargument name="key">
		<cfset var value = this.notfound>
		<cfif StructKeyExists(this.cache, key)>
			<cfset value = this.cache['#key#'].get()>
			<!---
			value may be null if it's been garbage collected,
			so check for this here.
			--->
			<cfif isDefined("value")>
				<cfset this.hits = this.hits + 1>
			<cfelse>
				<!---
				If value is null, we need to reap() to
				ensure that garbage collected references
				and keys pointing to them are cleaned out
				of the cache.
				--->
				<cfset this.misses = this.misses + 1>
				<cfset value = this.notfound>
				<cfset reap()>
			</cfif>
		<cfelse>
			<cfset this.misses = this.misses + 1>
		</cfif>
		<cfreturn value>
	</cffunction>
	
	
	<cffunction name="cacheSetElement" output="false" description="Puts a key/value pair into the cache">
		<cfargument name="key">
		<cfargument name="value">
		
		<!---
		This is Where the magic starts happening - wrap the value
		in a soft reference.
		--->
		<cfset var softRef = createObject("java", "java.lang.ref.SoftReference").init(value, this.refq)>
		
		<cfset this.cache['#key#'] = softRef>
		
		<!---
		We also maintain a reverse mapping From the soft reference
		to the key - as noted earlier, you'll see why in a minute...
		--->
		<cfset this.refKeyMap['#softRef#'] = key>
	</cffunction>
	
	
	<cffunction name="cacheRemoveElement" output="false" description="Deletes a key/value pair From the cache">
		<cfargument name="key">
		<cfset var softRef = "">
		<cfif StructKeyExists(this.cache, key)>
			<cfset softRef = this.cache['#key#']>
			
			<!--- clean it out From the main cache --->
			<cfset structDelete(this.cache, key)>
			<!--- and then From the soft reference to key map --->
			<cfset structDelete(this.refKeyMap, softRef)>
		</cfif>
	</cffunction>
	
	
	<cffunction name="reap" output="false" description="Reaps garbage collected values and keys From the cache. This is called
								automatically by the cache, but you may call it yourself if you really need to.">
		<cfset var more = true>
		<cfset var softRef = "">
		<cfset var key = "">
		<!--- 
		Check if a onRemoveHandler is defined. This is a CFC which is expected
		to implement an onRemove(key) function, which will be notified whenever
		a key is removed.
		--->
		<cfset var callHandler = isDefined("this.onRemoveHandler")>
		<cfloop condition="#more#">
			<!--- Poll the reference queue for garbage collected references --->
			<cfset softRef = this.refq.poll()>
			<!--- If we have a reference try to clean it up From the cache --->
			<cfif isDefined("softRef")>
				<!--- 
				We need to check whether it exists, since another thread
				may have already cleared it up. I think polling the queue
				is supposed to be atomic, but I've been running into a 
				little trouble here - in theory this check is not necessary.
				In practice... well...
				--->
				<cfif StructKeyExists(this.refKeyMap, softRef)>
					<!---
					Now you get to see why we need to refKeyMap - we need
					to get the key that the soft reference was mapped to.
					The value that the reference contained may be gone by
					now, so there's no way we could've stored the key
					in there along with the value. 
					--->
					<cfset key = this.refKeyMap['#softRef#']>
					<cfif callHandler>
						<!--- Call out to the onRemoveHandler, if we have one --->
						<cfset this.onRemoveHandler.onRemove(key)>
					</cfif>
					
					<!--- And last, but not least, clean up the cache! --->
					<cfset StructDelete(this.cache, key)>
					<cfset StructDelete(this.refKeyMap, softRef)>
				</cfif>
			<cfelse>
				<cfset more = false>
			</cfif>
		</cfloop>
	</cffunction>
	
	
	<cffunction name="getHitRatio" output="false" description="Returns the hit ratio for the cache">
		<cfset var requests = this.hits + this.misses>
		<cfif requests eq 0>
			<cfreturn 0>
		<cfelse>
			<cfreturn (this.hits/requests)>
		</cfif>
	</cffunction>
	
	
	<cffunction name="getSize" output="false" description="Returns the size of the cache">
		<!--- Reap first so we get dead keys out of the cache --->
		<cfset reap()>
		<cfreturn StructCount(this.cache)>
	</cffunction>
	
	<cffunction name="cacheShutdown" returntype="any" access="public" output="false">
		<cfargument name="key" type="string" required="true"/>
		<cfset StructClear(this.scope)>
	</cffunction>

</cfcomponent>